@@ -2,6 +2,7 @@
 import contextlib
 import itertools
 import logging
+from pathlib import Path
 import sys
 import time
 from dataclasses import dataclass
@@ -12,14 +13,13 @@ from typing import (
     Dict,
     Iterable,
     Iterator,
-    List,
     Optional,
     Tuple,
     Union,
     Generator,
     TextIO,
 )
-
+import csv
 import orjson
 import psycopg
 from orjson import JSONDecodeError
@@ -33,9 +33,23 @@ from tenacity import (
     wait_random_exponential,
 )
 
+
 from .db import PgstacDB
+from .hydration import dehydrate
 from enum import Enum
 
+logger = logging.getLogger(__name__)
+
+
+def chunked_iterable(iterable: Iterable, size: Optional[int] = 10000) -> Iterable:
+    """Chunk an iterable."""
+    it = iter(iterable)
+    while True:
+        chunk = tuple(itertools.islice(it, size))
+        if not chunk:
+            break
+        yield chunk
+
 
 class Tables(str, Enum):
     """Available tables for loading."""
@@ -89,52 +103,7 @@ def open_std(
                 pass
 
 
-def name_array_asdict(na: List) -> dict:
-    """Create a dict from a list with key from name field."""
-    out = {}
-    for i in na:
-        out[i["name"]] = i
-    return out
-
-
-def name_array_diff(a: List, b: List) -> List:
-    """Diff an array by name attribute."""
-    diff = dict_minus(name_array_asdict(a), name_array_asdict(b))
-    vals = diff.values()
-    return [v for v in vals if v != {}]
-
-
-def dict_minus(a: dict, b: dict) -> dict:
-    """Get a recursive difference between two dicts."""
-    out: dict = {}
-    for key, value in b.items():
-        if isinstance(value, list):
-            try:
-                arraydiff = name_array_diff(a[key], value)
-                if arraydiff is not None and arraydiff != []:
-                    out[key] = arraydiff
-                continue
-            except KeyError:
-                pass
-            except TypeError:
-                pass
-
-        if value is None or value == []:
-            continue
-        if a is None or key not in a:
-            out[key] = value
-            continue
-
-        if a.get(key) != value:
-            if isinstance(value, dict):
-                out[key] = dict_minus(a[key], value)
-                continue
-            out[key] = value
-
-    return out
-
-
-def read_json(file: Union[str, Iterator[Any]] = "stdin") -> Iterable:
+def read_json(file: Union[Path, str, Iterator[Any]] = "stdin") -> Iterable:
     """Load data from an ndjson or json file."""
     if file is None:
         file = "stdin"
@@ -148,9 +117,7 @@ def read_json(file: Union[str, Iterator[Any]] = "stdin") -> Iterable:
                     yield orjson.loads(line)
             except JSONDecodeError:
                 # If reading first line as json fails, try reading entire file
-                logging.info(
-                    "First line could not be parsed as json, trying full file."
-                )
+                logger.info("First line could not be parsed as json, trying full file.")
                 try:
                     f.seek(0)
                     json = orjson.loads(f.read())
@@ -160,7 +127,7 @@ def read_json(file: Union[str, Iterator[Any]] = "stdin") -> Iterable:
                     else:
                         yield json
                 except JSONDecodeError:
-                    logging.info("File cannot be read as json")
+                    logger.info("File cannot be read as json")
                     raise
     elif isinstance(file, Iterable):
         for line in file:
@@ -175,6 +142,7 @@ class Loader:
     """Utilities for loading data."""
 
     db: PgstacDB
+    _partition_cache: Optional[dict] = None
 
     @lru_cache
     def collection_json(self, collection_id: str) -> Tuple[dict, int, str]:
@@ -191,11 +159,12 @@ class Loader:
             raise Exception(
                 f"Collection {collection_id} is not present in the database"
             )
+        logger.debug(f"Found {collection_id} with base_item {base_item}")
         return base_item, key, partition_trunc
 
     def load_collections(
         self,
-        file: Union[str, Iterator[Any]] = "stdin",
+        file: Union[Path, str, Iterator[Any]] = "stdin",
         insert_mode: Optional[Methods] = Methods.insert,
     ) -> None:
         """Load a collections json or ndjson file."""
@@ -221,8 +190,8 @@ class Loader:
                         SELECT content FROM tmp_collections;
                         """
                     )
-                    logging.debug(cur.statusmessage)
-                    logging.debug(f"Rows affected: {cur.rowcount}")
+                    logger.debug(cur.statusmessage)
+                    logger.debug(f"Rows affected: {cur.rowcount}")
                 elif insert_mode in ("insert_ignore", "ignore"):
                     cur.execute(
                         """
@@ -231,8 +200,8 @@ class Loader:
                         ON CONFLICT DO NOTHING;
                         """
                     )
-                    logging.debug(cur.statusmessage)
-                    logging.debug(f"Rows affected: {cur.rowcount}")
+                    logger.debug(cur.statusmessage)
+                    logger.debug(f"Rows affected: {cur.rowcount}")
                 elif insert_mode == "upsert":
                     cur.execute(
                         """
@@ -242,8 +211,8 @@ class Loader:
                         UPDATE SET content=EXCLUDED.content;
                         """
                     )
-                    logging.debug(cur.statusmessage)
-                    logging.debug(f"Rows affected: {cur.rowcount}")
+                    logger.debug(cur.statusmessage)
+                    logger.debug(f"Rows affected: {cur.rowcount}")
                 else:
                     raise Exception(
                         "Available modes are insert, ignore, and upsert."
@@ -269,7 +238,7 @@ class Loader:
         conn = self.db.connect()
         t = time.perf_counter()
 
-        logging.debug(f"Loading data for partition: {partition}.")
+        logger.debug(f"Loading data for partition: {partition}.")
         with conn.cursor() as cur:
             with conn.transaction():
                 cur.execute(
@@ -296,7 +265,7 @@ class Loader:
                     ),
                 )
             with conn.transaction():
-                logging.debug(
+                logger.debug(
                     f"Adding partition {partition} took {time.perf_counter() - t}s"
                 )
                 t = time.perf_counter()
@@ -312,9 +281,18 @@ class Loader:
                     ) as copy:
                         for item in items:
                             item.pop("partition")
-                            copy.write_row(list(item.values()))
-                    logging.debug(cur.statusmessage)
-                    logging.debug(f"Rows affected: {cur.rowcount}")
+                            copy.write_row(
+                                (
+                                    item["id"],
+                                    item["collection"],
+                                    item["datetime"],
+                                    item["end_datetime"],
+                                    item["geometry"],
+                                    item["content"],
+                                )
+                            )
+                    logger.debug(cur.statusmessage)
+                    logger.debug(f"Rows affected: {cur.rowcount}")
                 elif insert_mode in (
                     "ignore_dupes",
                     "upsert",
@@ -337,9 +315,18 @@ class Loader:
                     ) as copy:
                         for item in items:
                             item.pop("partition")
-                            copy.write_row(list(item.values()))
-                    logging.debug(cur.statusmessage)
-                    logging.debug(f"Copied rows: {cur.rowcount}")
+                            copy.write_row(
+                                (
+                                    item["id"],
+                                    item["collection"],
+                                    item["datetime"],
+                                    item["end_datetime"],
+                                    item["geometry"],
+                                    item["content"],
+                                )
+                            )
+                    logger.debug(cur.statusmessage)
+                    logger.debug(f"Copied rows: {cur.rowcount}")
 
                     cur.execute(
                         sql.SQL(
@@ -358,8 +345,8 @@ class Loader:
                                 """
                             ).format(sql.Identifier(partition["partition"]))
                         )
-                        logging.debug(cur.statusmessage)
-                        logging.debug(f"Rows affected: {cur.rowcount}")
+                        logger.debug(cur.statusmessage)
+                        logger.debug(f"Rows affected: {cur.rowcount}")
                     elif insert_mode == "upsert":
                         cur.execute(
                             sql.SQL(
@@ -377,8 +364,8 @@ class Loader:
                             """
                             ).format(sql.Identifier(partition["partition"]))
                         )
-                        logging.debug(cur.statusmessage)
-                        logging.debug(f"Rows affected: {cur.rowcount}")
+                        logger.debug(cur.statusmessage)
+                        logger.debug(f"Rows affected: {cur.rowcount}")
                     elif insert_mode == "delsert":
                         cur.execute(
                             sql.SQL(
@@ -400,84 +387,123 @@ class Loader:
                             """
                             ).format(sql.Identifier(partition["partition"]))
                         )
-                        logging.debug(cur.statusmessage)
-                        logging.debug(f"Rows affected: {cur.rowcount}")
+                        logger.debug(cur.statusmessage)
+                        logger.debug(f"Rows affected: {cur.rowcount}")
                 else:
                     raise Exception(
                         "Available modes are insert, ignore, upsert, and delsert."
                         f"You entered {insert_mode}."
                     )
-        logging.debug(
+        logger.debug(
             f"Copying data for {partition} took {time.perf_counter() - t} seconds"
         )
 
+    def _partition_update(self, item: dict) -> str:
+
+        p = item.get("partition", None)
+        if p is None:
+            _, key, partition_trunc = self.collection_json(item["collection"])
+            if partition_trunc == "year":
+                pd = item["datetime"].replace("-", "")[:4]
+                p = f"_items_{key}_{pd}"
+            elif partition_trunc == "month":
+                pd = item["datetime"].replace("-", "")[:6]
+                p = f"_items_{key}_{pd}"
+            else:
+                p = f"_items_{key}"
+            item["partition"] = p
+
+        if self._partition_cache is None:
+            self._partition_cache = {}
+
+        partition = self._partition_cache.get(
+            item["partition"],
+            {
+                "partition": None,
+                "collection": None,
+                "mindt": None,
+                "maxdt": None,
+                "minedt": None,
+                "maxedt": None,
+            },
+        )
+
+        partition["partition"] = item["partition"]
+        partition["collection"] = item["collection"]
+        if partition["mindt"] is None or item["datetime"] < partition["mindt"]:
+            partition["mindt"] = item["datetime"]
+
+        if partition["maxdt"] is None or item["datetime"] > partition["maxdt"]:
+            partition["maxdt"] = item["datetime"]
+
+        if partition["minedt"] is None or item["end_datetime"] < partition["minedt"]:
+            partition["minedt"] = item["end_datetime"]
+
+        if partition["maxedt"] is None or item["end_datetime"] > partition["maxedt"]:
+            partition["maxedt"] = item["end_datetime"]
+        self._partition_cache[item["partition"]] = partition
+
+        return p
+
+    def read_dehydrated(self, file: Union[Path, str] = "stdin") -> Generator:
+        if file is None:
+            file = "stdin"
+        if isinstance(file, str):
+            open_file: Any = open_std(file, "r")
+            with open_file as f:
+                fields = [
+                    "id",
+                    "geometry",
+                    "collection",
+                    "datetime",
+                    "end_datetime",
+                    "content",
+                ]
+                csvreader = csv.DictReader(f, fields, delimiter="\t")
+                for item in csvreader:
+                    item["partition"] = self._partition_update(item)
+                    yield item
+
+    def read_hydrated(
+        self, file: Union[Path, str, Iterator[Any]] = "stdin"
+    ) -> Generator:
+        for line in read_json(file):
+            item = self.format_item(line)
+            item["partition"] = self._partition_update(item)
+            yield item
+
     def load_items(
         self,
-        file: Union[str, Iterator[Any]] = "stdin",
+        file: Union[Path, str, Iterator[Any]] = "stdin",
         insert_mode: Optional[Methods] = Methods.insert,
+        dehydrated: Optional[bool] = False,
+        chunksize: Optional[int] = 10000,
     ) -> None:
         """Load items json records."""
         if file is None:
             file = "stdin"
         t = time.perf_counter()
-        items: List = []
-        partitions: dict = {}
-        for line in read_json(file):
-            item = self.format_item(line)
-            items.append(item)
-            partition = partitions.get(
-                item["partition"],
-                {
-                    "partition": None,
-                    "collection": None,
-                    "mindt": None,
-                    "maxdt": None,
-                    "minedt": None,
-                    "maxedt": None,
-                },
-            )
-            partition["partition"] = item["partition"]
-            partition["collection"] = item["collection"]
-            if partition["mindt"] is None or item["datetime"] < partition["mindt"]:
-                partition["mindt"] = item["datetime"]
+        self._partition_cache = {}
 
-            if partition["maxdt"] is None or item["datetime"] > partition["maxdt"]:
-                partition["maxdt"] = item["datetime"]
+        if dehydrated and isinstance(file, str):
+            items = self.read_dehydrated(file)
+        else:
+            items = self.read_hydrated(file)
 
-            if (
-                partition["minedt"] is None
-                or item["end_datetime"] < partition["minedt"]
-            ):
-                partition["minedt"] = item["end_datetime"]
+        for chunk in chunked_iterable(items, chunksize):
+            list(chunk).sort(key=lambda x: x["partition"])
+            for k, g in itertools.groupby(chunk, lambda x: x["partition"]):
+                self.load_partition(self._partition_cache[k], g, insert_mode)
 
-            if (
-                partition["maxedt"] is None
-                or item["end_datetime"] > partition["maxedt"]
-            ):
-                partition["maxedt"] = item["end_datetime"]
-            partitions[item["partition"]] = partition
-        logging.debug(
-            f"Loading and parsing data took {time.perf_counter() - t} seconds."
-        )
-        t = time.perf_counter()
-        items.sort(key=lambda x: x["partition"])
-        logging.debug(f"Sorting data took {time.perf_counter() - t} seconds.")
-        t = time.perf_counter()
+        logger.debug(f"Adding data to database took {time.perf_counter() - t} seconds.")
 
-        for k, g in itertools.groupby(items, lambda x: x["partition"]):
-            self.load_partition(partitions[k], g, insert_mode)
-
-        logging.debug(
-            f"Adding data to database took {time.perf_counter() - t} seconds."
-        )
-
-    def format_item(self, _item: Union[str, dict]) -> dict:
+    def format_item(self, _item: Union[Path, str, dict]) -> dict:
         """Format an item to insert into a record."""
         out = {}
         item: dict
         if not isinstance(_item, dict):
             try:
-                item = orjson.loads(_item.replace("\\\\", "\\"))
+                item = orjson.loads(str(_item).replace("\\\\", "\\"))
             except:
                 raise Exception(f"Could not load {_item}")
         else:
@@ -521,12 +547,12 @@ class Loader:
         bbox = item.pop("bbox")
         geojson = item.pop("geometry")
         if geojson is None and bbox is not None:
-            pass
+            geometry = None
         else:
             geometry = str(Geometry.from_geojson(geojson).wkb)
         out["geometry"] = geometry
 
-        content = dict_minus(base_item, item)
+        content = dehydrate(base_item, item)
 
         out["content"] = orjson.dumps(content).decode()
 
